////////////////////////////////////////////////////////////////////// 
// Main.cpp 
// (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis
// 
// This file is distributed as part of the RareMetalWorker source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile RareMetalWorker.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday November 28, 2012
// 

#include "Parameters.h"
#include "Pedigree.h"
#include "math.h"
#include <iostream>
#include "FastFit.h"
#include "TraitTransformations.h"
#include "TransformResiduals.h"
#include <stdio.h>
#include <time.h>
#include "Error.h"
#include "AutoFit.h"
#include "PreMeta.h"
#include "OutputKin.h"
#include "WriteLog.h"
//#include <lapacke.h>
#include "DataQC.h"
#include "AutoFit2.h"

int main(int argc, char ** argv)
{
   printf("\nRAREMETALWORKER %s -- A Forerunner of RareMetal\n"
	 "          (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis\n\n", VERSION);

   String pedfile, datfile;
   double tol = 0.0001;

   BEGIN_LONG_PARAMETERS(additional)
      LONG_PARAMETER_GROUP("Input Files")
      LONG_STRINGPARAMETER("ped", &pedfile)
      LONG_STRINGPARAMETER("dat", &datfile)
      LONG_STRINGPARAMETER("vcf", &PreMeta::vcfInput)
      LONG_PARAMETER_GROUP("Output Files")
      LONG_STRINGPARAMETER("prefix", &PreMeta::outputFile)
      LONG_INTPARAMETER("LDwindow", &PreMeta::window)
      LONG_PARAMETER_GROUP("VC Options")
      //LONG_PARAMETER("vcShared", &AutoFit::fitSharedEnvironment)
      LONG_PARAMETER("vcX", &AutoFit::fitX)
      LONG_PARAMETER("useCovariates", &FastFit::useCovariates)
      LONG_PARAMETER_GROUP("Trait Options")
      LONG_PARAMETER("makeResiduals", &FastFit::makeResiduals)
      LONG_PARAMETER("inverseNormal", &FastFit::inverseNormal)
      LONG_STRINGPARAMETER("traitName", &FastFit::traitName)
      LONG_PARAMETER_GROUP("Kinship Source")
      LONG_PARAMETER("kinPedigree", &FastTransform::pedKin)
      LONG_PARAMETER("kinGeno", &FastTransform::empKin)
      LONG_STRINGPARAMETER("kinFile", &FastTransform::readInEmp)
      LONG_PARAMETER("kinSave", &OutputKin::outputKin)
      LONG_PARAMETER_GROUP("Kinship Options")
      LONG_DOUBLEPARAMETER("kinMaf", &KinshipEmp::q)
      LONG_DOUBLEPARAMETER("kinMiss", &KinshipEmp::miss)
      LONG_PARAMETER_GROUP("Chromosome X")
      LONG_STRINGPARAMETER("xLabel", &PreMeta::xLabel)
      LONG_INTPARAMETER("xStart", &PreMeta::Xstart)
      LONG_INTPARAMETER("xEnd", &PreMeta::Xend)
      LONG_INTPARAMETER("maleLabel", &PreMeta::maleLabel)
      LONG_INTPARAMETER("femaleLabel", &PreMeta::femaleLabel)

      END_LONG_PARAMETERS();


   ParameterList pl;
   pl.Add(new LongParameters("Options:", additional));
   pl.Read(argc, argv);

   String logFile;
   if(PreMeta::outputFile != "")
   {
      if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	 logFile = PreMeta::outputFile + "singlevar.log";
      else
	 logFile = PreMeta::outputFile + ".singlevar.log";
   }
   else
      logFile = "singlevar.log";

   FILE * log;

   pl.Status();
   log = freopen (logFile,"wt",stderr);

   Pedigree ped;

   printf("Loading PED/DAT files ...\n");
   fprintf(log,"Loading input files ...\n");
   printf("  Loading DAT files ...");
   fprintf(log,"  Loading DAT files ...");
   ped.Prepare(datfile);
   printf("done.\n  Loading PED files ...");
   fprintf(log,"done.\n  Loading PED files ...");
   ped.Load(pedfile);
   printf("done.\n\n");
   fprintf(log,"done.\n\n");

   SanityCheck checkData;
   checkData.Check(ped,log);

   FastFit engine;
   FastTransform trans;
   KinshipEmp kin_emp;

   //check if the sample IDs from PED and VCF files are consistent
   trans.ScreenSampleID(ped,FastFit::useCovariates);

   //setup empirical kinship if applicable
   if(FastTransform::empKin)
      kin_emp.SetupEmpKin(ped,trans.genotypedSamplePED,trans.genotypedSampleVCF,trans.samplePEDIDHash,checkData.skippedSNPs,log);

   StringArray traits;

   if(!FastFit::traitName.IsEmpty())
   {
      FILE * trFile = fopen(FastFit::traitName,"r");
      if(trFile!=NULL)
      {
	 String buffer;
	 while (!feof(trFile))
	 {
	    buffer.ReadLine(trFile);
	    traits.AddTokens(buffer, SEPARATORS);
	 }
	 fclose(trFile);
      }
      else
      {
	 traits.AddTokens(FastFit::traitName,"/");
      }
   }
   else 
   {
      traits = ped.traitNames;
   }

   for(int i=0;i<traits.Length();i++)
   {
      FastFit::traitName = traits[i];
      printf("Analyzing trait %s\n",traits[i].c_str());
      fprintf(log,"Analyzing trait %s\n",traits[i].c_str());

      FILE * SCOREoutput;
      FILE * SCOREcov;
      String filename,filename_COV;

      if(PreMeta::outputFile != "")
      {
	 if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	 {
	    filename = PreMeta::outputFile + traits[i] + ".singlevar.score.txt";
	    filename_COV = PreMeta::outputFile + traits[i] + ".singlevar.cov.txt";
	 }
	 else
	 {
	    filename = PreMeta::outputFile + "." + traits[i] + ".singlevar.score.txt";
	    filename_COV = PreMeta::outputFile + "." + traits[i] + ".singlevar.cov.txt";
	 }
      }
      else
      {
	 filename_COV = traits[i] + ".singlevar.cov.txt";
	 filename = traits[i] + ".singlevar.score.txt";
      }

      SCOREoutput = fopen(filename,"wt");
      SCOREcov = fopen(filename_COV,"wt");

      engine.FitModels(SCOREoutput,ped,tol,trans,kin_emp,log);

      PreMeta preMeta;
      preMeta.Run(SCOREoutput,SCOREcov,ped,trans,engine,log);

      //printf("#GENOMIC CONTROL LAMBDA = %g\n",preMeta.lambda);

      fclose(SCOREoutput);
      fclose(SCOREcov);

      printf("completed.\n");
      fprintf(log,"completed.\n");
      String scoreFile,covFile;
      printf("  Single variant statistics stored in %s.\n",filename.c_str());
      fprintf(log,"  Single variant statistics stored in %s.\n",filename.c_str());

      printf("  Variance-covariance matrix stored in %s.\n\n",filename_COV.c_str());
      fprintf(log,"  Variance-covariance matrix stored in %s.\n\n",filename_COV.c_str());
   }
   WriteLog(pedfile,datfile,log);
   printf("Log file listing current options stored in %s\n\n",logFile.c_str());
   fprintf(log,"Log file listing current options stored in %s\n\n",logFile.c_str());

   //fprintf(log,"\nlog file has been saved in %s\n\n",logFile.c_str());

   fclose(log);
   //fclose(stderr);
   return 0;
}
